/* 
pantelmail copyright Jan Panteltje 2008-always
email: panteltje@yahoo.com

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "pantelmail.h"


FILE *header_filefd;/* must be preserved between calls */
FILE *body_filefd;
FILE *custom_headers_filefd;
FILE *attachment_filefd;

time_t last_body_line_time;
time_t now;
int mail_mode;


/*         
211 System status, or system help reply
214 Help message
     [Information on how to use the receiver or the meaning of a
     particular non-standard command; this reply is useful only
     to the human user]
220 <domain> Service ready
221 <domain> Service closing transmission channel
250 Requested mail action okay, completed
251 User not local; will forward to <forward-path>
 
354 Start mail input; end with <CRLF>.<CRLF>
 
421 <domain> Service not available,
     closing transmission channel
     [This may be a reply to any command if the service knows it
     must shut down]
450 Requested mail action not taken: mailbox unavailable
     [E.g., mailbox busy]
451 Requested action aborted: local error in processing
452 Requested action not taken: insufficient system storage
 
500 Syntax error, command unrecognized
     [This may include errors such as command line too long]
501 Syntax error in parameters or arguments
502 Command not implemented
503 Bad sequence of commands
504 Command parameter not implemented
550 Requested action not taken: mailbox unavailable
     [E.g., mailbox not found, no access]
551 User not local; please try <forward-path>
552 Requested mail action aborted: exceeded storage allocation
553 Requested action not taken: mailbox name not allowed
     [E.g., mailbox syntax incorrect]
554 Transaction failed
*/


int connect_to_mail_server(char *mail_server, int port)
{
struct hostent *hp;
struct sockaddr_in sa; /* -Wall gets confused */
int a, i;
char temp[TEMP_SIZE];
char dotted_quad[TEMP_SIZE];
int flags;
time_t connect_timer;
extern int idle_cb();
int connect_flag;
unsigned long in;
int byname;

if(debug_flag)
	{
	fprintf(stderr, "connect_to_mail_server(): arg mail_server=%s port=%d\n", mail_server, port);
	}
	
/* argument check */
if(! mail_server) return 0;
if(port <= 0) return 0;

mail_server_status = DISCONNECTED;

/* test for name or address */
hp = 0;
in = inet_addr (mail_server);
if(in != INADDR_NONE) /* address */
	{
	byname = 0;
	if(debug_flag)
		{
		fprintf(stderr, "getting host %s by address", mail_server);
		}

	hp = gethostbyaddr((char *)&in, sizeof(in), AF_INET);
	if(! hp)
		{
		if(debug_flag)
			{
			fprintf(stderr, "gethostbyaddress failed: cannot get host %s by address", mail_server);
			}
		}
	}
else /* name */
	{
	byname = 1;
	if(debug_flag)
		{
		fprintf(stderr, "getting host %s by name\n", mail_server);
		}

	hp = gethostbyname(mail_server);
	if(! hp)
		{
		fprintf(stderr, "gethostbyname failed: cannot get host %s by name\n", mail_server);
		}
	}

if(! hp)
	{
	/* 
	make sure if running unattended that program does not wait for user
	to press OK in fl_show_alert, since the dialup would never disconnect.
	*/

	if(! auto_disconnect_from_net_flag)
		{
//		an_idle_cb = fl_set_idle_callback(0, 0);
//		fl_show_alert(temp, "", "", 0);

		if(byname)
			{
//			fl_show_alert("gethostbyname", mail_server, "not found", 0);
			}
		else
			{
//			fl_show_alert("gethostbyaddress", mail_server, "not found", 0);
			}

//		fl_set_idle_callback(idle_cb, an_idle_cb);
		}
	else
		{
		/* NO WAIT */
//		to_error_log(temp);
//		to_command_status_display(temp);
		}

	return 0;
	}

/*
 - Data Type: struct hostent
     This data type is used to represent an entry in the hosts
     database.  It has the following members:

    `char *h_name'
          This is the "official" name of the host.

    `char **h_aliases'
          These are alternative names for the host, represented as a
          null-terminated vector of strings.

    `int h_addrtype'
          This is the host address type; in practice, its value is
          always either `AF_INET' or `AF_INET6', with the latter being
          used for IPv6 hosts.  In principle other kinds of addresses
          could be represented in the database as well as Internet
          addresses; if this were done, you might find a value in this
          field other than `AF_INET' or `AF_INET6'.  *Note Socket
          Addresses::.

    `int h_length'
          This is the length, in bytes, of each address.

    `char **h_addr_list'
          This is the vector of addresses for the host.  (Recall that
          the host might be connected to multiple networks and have
          different addresses on each one.)  The vector is terminated
          by a null pointer.

    `char *h_addr'
          This is a synonym for `h_addr_list[0]'; in other words, it is
          the first host address.

*/

#ifdef SHOW_ALIASES
/* char **h_aliases nul terminated vector of strings */

char **ptr;
ptr = hp -> h_aliases;
while(1)
	{
	if(*ptr == 0) break;	

	
	fprintf(stderr, "%s\n", *ptr);	

	**ptr++;	
	
	}
//fprintf(stderr, "done.\n");
//exit(1);
#endif // SHOW_ALIASES


/* gethostbyname() leaves port and host address in network byte order */
bzero(&sa, sizeof(sa) );
bcopy(hp -> h_addr, (char *)&sa.sin_addr, hp -> h_length);
sa.sin_family = AF_INET;
sa.sin_port = htons( (u_short)port);

/* sa.sin_addr and sa.sin_port now in network byte order */

/* create a socket */
mail_socketfd = socket(hp -> h_addrtype, SOCK_STREAM, 0);
if(mail_socketfd < 0)
	{
	sprintf(temp, "Cannot create socket");
	if(debug_flag) fprintf(stderr, "%s\n", temp);
				
	return 0;
	}/* end mail_socketfd < 0 */

//sprintf(dotted_quad, "%s", (char *)inet_ntoa( (struct in_addr)sa.sin_addr.s_addr) );
sprintf(dotted_quad, "%s", (char *)inet_ntoa(sa.sin_addr.s_addr) );

//fprintf(stderr, "pantelmail: IP=%s\n", dotted_quad);
//exit(1);

/* set for nonblocking socket */
if (fcntl(mail_socketfd, F_SETFL, O_NONBLOCK) < 0)
	{
	sprintf(temp, "Cannot set socket non blocking");
	if(debug_flag) fprintf(stderr, "%s\n", temp);
				
	if(! auto_disconnect_from_net_flag)
		{
//		an_idle_cb = fl_set_idle_callback(0, 0);
//		fl_show_alert(temp, "", "", 0);
//		fl_set_idle_callback(idle_cb, an_idle_cb);
		}
	else
		{
		/* NO WAIT */
//		to_error_log(temp);
//		to_command_status_display(temp);
		}
	return 0;
	}

/* connect */
/*
sprintf(temp, "Connecting to mail server %s (%s)  port %d  timeout %d",\
mail_server, dotted_quad, port, connect_to_mail_server_timeout);
if(debug_flag)
	{
	fprintf(stderr, "%s\n", temp);
	}
to_command_status_display(temp);
*/

/* mail_server sometimes does not seem to answere to the first connect */
/* not so sure this is the way to do it, may change this later */
for(i = 0; i < MAIL_SERVER_CONNECT_RETRIES; i++)
	{
	fprintf(stderr,\
	"pantelmail: connecting to mail server %s (%s)  port %d  timeout %d  retry %d (of %d)\n",\
	mail_server, dotted_quad, port, connect_to_mail_server_timeout, i + 1, MAIL_SERVER_CONNECT_RETRIES);

	/* start the connect timer */
	connect_flag = 0;
	connect_timer = time(0);
	while(1)
		{
		a = connect(mail_socketfd, (struct sockaddr*)&sa, sizeof(sa) );
		if(a == 0)
			{
			connect_flag = 1;
			break;/* connected */
			}

		if(debug_flag)
			{
//			fprintf(stderr, "In connect errno=%d\n", errno);

//			fprintf(stderr ,"pantelmail: connect failed because: ");
//			perror("");
			}

		if(errno == 111)
			{
			fprintf(stderr, "pantelmail: could not send mail, because: \n");
			perror("");
			exit(1);
			}
	
		/* test for connect time out */
		if( (time(0) - connect_timer) > connect_to_mail_server_timeout)
			{
			fprintf(stderr, "pantelmail: connect timeout %s\n", temp);
				
			fprintf(stderr, "pantelmail: could not send mail.\n");

exit(1);
			/* try again */
//			break;
		
			/* close the socket */
			close(mail_socketfd);
		
			return 0;
			}/* end timeout */
		}/* end while connect */

	if(connect_flag) break;

	}/* end for */

if(! connect_flag)
	{
	/* close the socket */
	close(mail_socketfd);

//	to_error_log("Connect to mail_server failed");

	return 0;
	}

/* now set blocking */
flags = fcntl(mail_socketfd, F_GETFL, 0);
flags &= ~O_NONBLOCK;
fcntl(mail_socketfd, F_SETFL, flags);

sprintf(temp, "Connected to mail server %s (%s)  port %d",\
mail_server, dotted_quad, port);
if(debug_flag)
	{
	fprintf(stderr, "%s\n", temp);
	}
//to_command_status_display(temp);

mail_server_status = CONNECTED;

return 1;
}/* end function connect_to_mail_server */


int send_to_mail_server(char *text)/* send data to mail_server */
{
int a, c, i, j;
char txbuf[READSIZE];

if(debug_flag)
	{
	fprintf(stderr, "send_to_mail_server(): arg\ntext=%s (mail_socketfd=%d)\n", text, mail_socketfd);
	}
	
/* argument check */
if(! text) return 0;

if(mail_server_status == DISCONNECTED) return 0;

/* expand any LF ('\n' 10 decimal) to CRLF (13 10) */
i = 0;
j = 0;
while(1)
	{
	c = text[i];
	if(c == '\n')
		{
		txbuf[j] = 13;
		j++;
		if(j > READSIZE)
			{
			if(debug_flag)
				{
				fprintf(stderr,\
				"send_to_mail_server(): overflow in expanding text=%s\n", text);
				}
			return 0;
			}		
		}
	txbuf[j] = c;
	if(! c) break;
	i++;
	j++;
	}

a = write(mail_socketfd, txbuf, strlen(txbuf) );     
if(a < 0)
	{
	if(debug_flag)
		{
		fprintf(stderr, "\nwrite failed, errno=%d\n", errno);
		}
	perror("send_to_mail_server(): write failed:");
	return 0;
	}

/* update bytes_send */
bytes_send += (long) strlen(txbuf);

if(debug_flag)
	{
	fprintf(stderr, "txbuf=%s", txbuf);
	}

return 1;
}/* end function send_to_mail_server */


