/* 
pantelmail copyright Jan Panteltje 2008-always
email: panteltje@yahoo.com

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/




#include "pantelmail.h"


fd_set readfs;	/* data type represents file descriptor for select function, (actually it is a bit array) */
struct timeval timeout;
char domainname[4096];



char *strsave(char *s) /* save char array s somewhere */
{
char *p;

p = malloc(strlen(s) +  1);
if(p) strcpy(p, s);
return p;
} /* end function strsave */



int readline(FILE *file, char *contents)
{
int a, c, i;

for(i = 0; i < READSIZE - 1; i++)
	{
	while(1)
		{
		c = getc(file);
		a = ferror(file);
		if(a)
			{
			perror("readline():");
			continue;
			}
		break;
		}
	if(feof(file) )
		{
		fclose(file);
		contents[i] = 0;/* EOF marker */
		return EOF;
		}
	if(c == '\n')
		{
		contents[i] = 0;/* string termination */
		return 1;/* end of line */
		} 
	contents[i] = c;
	}
/*
mmm since we are here, the line must be quite long, possibly something
is wrong.
Since I do not always check for a return 0 in the use uf this function,
just to be safe, gona force a string termination.
This prevents stack overflow, and variables getting overwritten.
*/
contents[i] = 0;/* force string termination */
/*fclose(file);*/
if(debug_flag)
	{
	fprintf(stderr,\
	"readline(): line to long, returning 0 contents=%s\n", contents);
	}
return 0;
}/* end function readline */


int connect_to_mail_server(char *mail_server, int port)
{
struct hostent *hp;
struct sockaddr_in sa; /* -Wall gets confused */
int a, i;
char temp[TEMP_SIZE];
char dotted_quad[TEMP_SIZE];
int flags;
time_t connect_timer;
extern int idle_cb();
int connect_flag;
unsigned long in;
int byname;

if(debug_flag)
	{
	fprintf(stderr, "connect_to_mail_server(): arg mail_server=%s port=%d\n", mail_server, port);
	}
	
/* argument check */
if(! mail_server) return 0;
if(port <= 0) return 0;

mail_server_status = DISCONNECTED;

/* test for name or address */
hp = 0;
in = inet_addr (mail_server);
if(in != INADDR_NONE) /* address */
	{
	byname = 0;
	if(debug_flag)
		{
		fprintf(stderr, "getting host %s by address", mail_server);
		}

	hp = gethostbyaddr((char *)&in, sizeof(in), AF_INET);
	if(! hp)
		{
		if(debug_flag)
			{
			fprintf(stderr, "gethostbyaddress failed: cannot get host %s by address", mail_server);
			}
		}
	}
else /* name */
	{
	byname = 1;
	if(debug_flag)
		{
		fprintf(stderr, "getting host %s by name\n", mail_server);
		}

	hp = gethostbyname(mail_server);
	if(! hp)
		{
		fprintf(stderr, "gethostbyname failed: cannot get host %s by name\n", mail_server);
		}
	}

if(! hp)
	{
	/* 
	make sure if running unattended that program does not wait for user
	to press OK in fl_show_alert, since the dialup would never disconnect.
	*/

	if(! auto_disconnect_from_net_flag)
		{
//		an_idle_cb = fl_set_idle_callback(0, 0);
//		fl_show_alert(temp, "", "", 0);

		if(byname)
			{
//			fl_show_alert("gethostbyname", mail_server, "not found", 0);
			}
		else
			{
//			fl_show_alert("gethostbyaddress", mail_server, "not found", 0);
			}

//		fl_set_idle_callback(idle_cb, an_idle_cb);
		}
	else
		{
		/* NO WAIT */
//		to_error_log(temp);
//		to_command_status_display(temp);
		}

	return 0;
	}

/*
 - Data Type: struct hostent
     This data type is used to represent an entry in the hosts
     database.  It has the following members:

    `char *h_name'
          This is the "official" name of the host.

    `char **h_aliases'
          These are alternative names for the host, represented as a
          null-terminated vector of strings.

    `int h_addrtype'
          This is the host address type; in practice, its value is
          always either `AF_INET' or `AF_INET6', with the latter being
          used for IPv6 hosts.  In principle other kinds of addresses
          could be represented in the database as well as Internet
          addresses; if this were done, you might find a value in this
          field other than `AF_INET' or `AF_INET6'.  *Note Socket
          Addresses::.

    `int h_length'
          This is the length, in bytes, of each address.

    `char **h_addr_list'
          This is the vector of addresses for the host.  (Recall that
          the host might be connected to multiple networks and have
          different addresses on each one.)  The vector is terminated
          by a null pointer.

    `char *h_addr'
          This is a synonym for `h_addr_list[0]'; in other words, it is
          the first host address.

*/

#ifdef SHOW_ALIASES
/* char **h_aliases nul terminated vector of strings */

char **ptr;
ptr = hp -> h_aliases;
while(1)
	{
	if(*ptr == 0) break;	

	
	fprintf(stderr, "%s\n", *ptr);	

	**ptr++;	
	
	}
//fprintf(stderr, "done.\n");
//exit(1);
#endif // SHOW_ALIASES


/* gethostbyname() leaves port and host address in network byte order */
bzero(&sa, sizeof(sa) );
bcopy(hp -> h_addr, (char *)&sa.sin_addr, hp -> h_length);
sa.sin_family = AF_INET;
sa.sin_port = htons( (u_short)port);

/* sa.sin_addr and sa.sin_port now in network byte order */

/* create a socket */
mail_socketfd = socket(hp -> h_addrtype, SOCK_STREAM, 0);
if(mail_socketfd < 0)
	{
	sprintf(temp, "Cannot create socket");
	if(debug_flag) fprintf(stderr, "%s\n", temp);
				
	return 0;
	}/* end mail_socketfd < 0 */

//sprintf(dotted_quad, "%s", (char *)inet_ntoa( (struct in_addr)sa.sin_addr.s_addr) );
sprintf(dotted_quad, "%s", (char *)inet_ntoa(sa.sin_addr.s_addr) );

//fprintf(stderr, "pantelmail: IP=%s\n", dotted_quad);
//exit(1);

/* set for nonblocking socket */
if (fcntl(mail_socketfd, F_SETFL, O_NONBLOCK) < 0)
	{
	sprintf(temp, "Cannot set socket non blocking");
	if(debug_flag) fprintf(stderr, "%s\n", temp);
				
	if(! auto_disconnect_from_net_flag)
		{
//		an_idle_cb = fl_set_idle_callback(0, 0);
//		fl_show_alert(temp, "", "", 0);
//		fl_set_idle_callback(idle_cb, an_idle_cb);
		}
	else
		{
		/* NO WAIT */
//		to_error_log(temp);
//		to_command_status_display(temp);
		}
	return 0;
	}

/* connect */
/*
sprintf(temp, "Connecting to mail server %s (%s)  port %d  timeout %d",\
mail_server, dotted_quad, port, connect_to_mail_server_timeout);
if(debug_flag)
	{
	fprintf(stderr, "%s\n", temp);
	}
to_command_status_display(temp);
*/

/* mail_server sometimes does not seem to answere to the first connect */
/* not so sure this is the way to do it, may change this later */
for(i = 0; i < MAIL_SERVER_CONNECT_RETRIES; i++)
	{
	fprintf(stderr,\
	"pantelmail: connecting to mail server %s (%s)  port %d  timeout %d  retry %d (of %d)\n",\
	mail_server, dotted_quad, port, connect_to_mail_server_timeout, i + 1, MAIL_SERVER_CONNECT_RETRIES);

	/* start the connect timer */
	connect_flag = 0;
	connect_timer = time(0);
	while(1)
		{
		a = connect(mail_socketfd, (struct sockaddr*)&sa, sizeof(sa) );
		if(a == 0)
			{
			connect_flag = 1;
			break;/* connected */
			}

		if(debug_flag)
			{
//			fprintf(stderr, "In connect errno=%d\n", errno);

//			fprintf(stderr ,"pantelmail: connect failed because: ");
//			perror("");
			}

		if(errno == 111)
			{
			fprintf(stderr, "pantelmail: could not send mail, because: \n");
			perror("");
			exit(1);
			}
	
		/* test for connect time out */
		if( (time(0) - connect_timer) > connect_to_mail_server_timeout)
			{
			fprintf(stderr, "pantelmail: connect timeout %s\n", temp);
				
			fprintf(stderr, "pantelmail: could not send mail.\n");

exit(1);
			/* try again */
//			break;
		
			/* close the socket */
			close(mail_socketfd);
		
			return 0;
			}/* end timeout */
		}/* end while connect */

	if(connect_flag) break;

	}/* end for */

if(! connect_flag)
	{
	/* close the socket */
	close(mail_socketfd);

//	to_error_log("Connect to mail_server failed");

	return 0;
	}

/* now set blocking */
flags = fcntl(mail_socketfd, F_GETFL, 0);
flags &= ~O_NONBLOCK;
fcntl(mail_socketfd, F_SETFL, flags);

sprintf(temp, "Connected to mail server %s (%s)  port %d",\
mail_server, dotted_quad, port);
if(debug_flag)
	{
	fprintf(stderr, "%s\n", temp);
	}
//to_command_status_display(temp);

mail_server_status = CONNECTED;

return 1;
}/* end function connect_to_mail_server */


int send_to_mail_server(char *text)/* send data to mail_server */
{
int a, c, i, j;
char txbuf[READSIZE];

if(debug_flag)
	{
	fprintf(stderr, "send_to_mail_server(): arg\ntext=%s (mail_socketfd=%d)\n", text, mail_socketfd);
	}
	
/* argument check */
if(! text) return 0;

if(mail_server_status == DISCONNECTED) return 0;

/* expand any LF ('\n' 10 decimal) to CRLF (13 10) */
i = 0;
j = 0;
while(1)
	{
	c = text[i];
	if(c == '\n')
		{
		txbuf[j] = 13;
		j++;
		if(j > READSIZE)
			{
			if(debug_flag)
				{
				fprintf(stderr,\
				"send_to_mail_server(): overflow in expanding text=%s\n", text);
				}
			return 0;
			}		
		}
	txbuf[j] = c;
	if(! c) break;
	i++;
	j++;
	}

a = write(mail_socketfd, txbuf, strlen(txbuf) );     
if(a < 0)
	{
	if(debug_flag)
		{
		fprintf(stderr, "\nwrite failed, errno=%d\n", errno);
		}
	perror("send_to_mail_server(): write failed:");
	return 0;
	}

/* update bytes_send */
bytes_send += (long) strlen(txbuf);

if(debug_flag)
	{
	fprintf(stderr, "txbuf=%s", txbuf);
	}

return 1;
}/* end function send_to_mail_server */


#ifdef OLD_GET_UNIVERSAL_TIME_AS_STRING
char *get_universal_time_as_string()
{
time_t now;
static char temp[TEMP_SIZE];
size_t strftime();
struct tm *universal_time, *gmtime();
char *ptr;
char *strsave();
char *old_locale, *saved_locale;
int locale_modified_flag;


if(debug_flag)
	{
	fprintf(stderr, "get_universal_time_as_string(): arg none\n");
	}


locale_modified_flag = 0;

/* 
set the locale to english language even if env shows otherwise
*/
/* Get the name of the current locale */
old_locale = setlocale (LC_ALL, NULL);
if(! old_locale)
	{
	fprintf(stderr, "get_universal_time_as_string(): could not get current locale, posting not possible\n");
	
	return 0;
	}

/* Copy the name so it won't be clobbered by `setlocale' */
saved_locale = strdup (old_locale);
if(saved_locale == NULL)
	{
	fprintf(stderr, "get_universal_time_as_string(): Out of memory\n");

	return 0;
	}     

/* 
Now change the locale to US english for 'Wed and not for example 'wo'
server may not be able to parse something like 'wo'.
*/
ptr = setlocale(LC_ALL, THE_LOCALE);
if(! ptr)
	{
	fprintf(stderr,\
	"get_universal_time_as_string(); could not setlocale,\n\
	perhaps run 'locale -a' and modify THE_LOCALE in NewsFleX.h?,\n\
	aborting posting!\n");

//	return 0;

//exit(1);
	}
else
	{
	locale_modified_flag = 1;
	}

/* 
decided to adapt to universal time (UT == GMT).
New format:
from rfc1036.html:
Wdy, DD Mon YY HH:MM:SS TIMEZONE 
*/

now = time(0);
universal_time = gmtime(&now);
strftime (temp, 511, "%a, %d %b %Y %H:%M:%S GMT", universal_time);

ptr = strsave(temp);

if(debug_flag)
	{
	fprintf(stderr, "get_universal_time_as_string(): ptr=%s\n", ptr);
	}

if(locale_modified_flag)
	{
	/* Restore the original locale */
	if(! setlocale (LC_ALL, saved_locale) )
		{
		fprintf(stderr, "get_universal_time_as_string() restetting original locale failed, posting impossible\n");

//		return 0;
		}
	}

free (saved_locale);

return ptr;
}/* end function get_universal_time_as_string */

#else

char *get_universal_time_as_string()
{
time_t now;
static char temp[TEMP_SIZE];
size_t strftime();
struct tm *universal_time, *gmtime();
char *ptr;
char *strsave();
char *old_locale, *saved_locale;
int locale_modified_flag;

if(debug_flag)
    {
    fprintf(stderr, "get_universal_time_as_string(): arg none\n");
    }

locale_modified_flag = 0;

/* 
set the locale to english language even if env shows otherwise
*/
/* Get the name of the current locale */
old_locale = setlocale (LC_ALL, NULL);
if(! old_locale)
    {
    fprintf(stderr, "get_universal_time_as_string(): could not get current locale, posting not possible\n");

    return 0;
    }

/* Copy the name so it won't be clobbered by `setlocale' */
saved_locale = strdup (old_locale);
if(saved_locale == NULL)
    {
    fprintf(stderr, "get_universal_time_as_string(): Out of memory\n");

    return 0;
    }     

/* 
Now change the locale to US english for 'Wed and not for example 'wo'
server may not be able to parse something like 'wo'.
*/
ptr = setlocale(LC_ALL, "en_US");
if(! ptr)
    {
    fprintf(stderr,\
    "get_universal_time_as_string(); could not setlocale,\n\
    perhaps run 'locale -a' and modify the locale to POSIX?\n");

//  return 0;
    }
else
    {
    locale_modified_flag = 1;
    }

/* 
decided to adapt to universal time (UT == GMT).
New format:
from rfc1036.html:
Wdy, DD Mon YY HH:MM:SS TIMEZONE 
*/

now = time(0);
universal_time = gmtime(&now);
strftime (temp, 511, "%a, %d %b %Y %H:%M:%S GMT", universal_time);

ptr = strsave(temp);

if(debug_flag)
    {
    fprintf(stderr, "get_universal_time_as_string(): ptr=%s\n", ptr);
    }

if(locale_modified_flag)
    {
    /* Restore the original locale */
    if(! setlocale (LC_ALL, saved_locale) )
       {
       fprintf(stderr, "get_universal_time_as_string() resetting original locale failed.\n");

//     return 0;
       }
   }

free (saved_locale);

return ptr;
}/* end function get_universal_time_as_string */

#endif


int print_usage()
{
fprintf(stderr, "\
Usage:\n\
pantelmail -a subject [-d] -f from [-h] [-p port] -r recipient [-s servername] -t textfilename\n\n\
-a subject           subject, example\n\
-d                   debug mode, prints functions and arguments, and server messages.\n\
-f from              from, example: \"Name Surname <me@mydomain.com>\", include the quotes.\n\
-h                   help, this help.\n\
-p n                 port, default 25.\n\
-r recipient         recipient, example: \"Name Surname <he@hisdomain.com>\", include the quotes.\n\
-s servername        use servername, not domain of recipient.\n\
-t textfilename      textfilename.\n\
Example:\n\
      pantelmail -a \"my subject\" -t my_text_file.txt -f \"Me Myself <me@myhost>\" -r \"You Inperson <you@yourhost>\"\n\
");


return 1;
} /* end function print_usage */



int main(int argc, char **argv)
{
int a, b;
char temp[4096];
size_t len = 4095;
int error_flag;

/* get user info */
userinfo = getpwuid(getuid() );
user_name = strsave(userinfo -> pw_name);
home_dir = strsave(userinfo -> pw_dir);
if(! home_dir)
	{
	fprintf(stderr, "pantelmail: main: could not allocate space for home_dir, aborting.\n");
	
	exit(1);
	}

/* defaults */
mail_server_port = 25;

connect_to_mail_server_timeout = 10;

//int getdomainname(char *name, size_t len);
if( getdomainname(domainname, len) < 0)
	{
	fprintf(stderr, "pantelmail: could not get domainname, because: \n");
	perror("");
	fprintf(stderr, "pantelmail: aborting.\n");
	
	exit(1);
	}

sprintf(temp, " <%s@%s>", user_name, domainname);
sender = strsave(temp);
if(! sender)
	{
	fprintf(stderr, "pantelmail: could not allocate space for sender, aborting.\n");
	
	exit(1);
	}

recipient = strsave("");
if(! recipient)
	{
	fprintf(stderr, "pantelmail: could not allocate space for recipient, aborting.\n");
	
	exit(1);
	}

text_filename = strsave("");
if(! text_filename)
	{
	fprintf(stderr, "pantelmail: could not allocate space for text_filename, aborting.\n");
	
	exit(1);
	}

subject = strsave("");
if(! subject)
	{
	fprintf(stderr, "pantelmail: could not allocate space for subject, aborting.\n");

	exit(1);	
	}

mail_server_name = strsave("");
if(! mail_server_name)
	{
	fprintf(stderr, "pantelmail: could not allocate space for mail_server_name, aborting.\n");

	exit(1);
	}

/* end defaults */

/* get command line options */
while(1)
	{
	a = getopt(argc, argv, "a:df:hp:r:s:t:");
	if(a == -1) break;

	switch(a)
		{
		case 'a':
			free(subject);
			subject = strsave(optarg);
			if(! subject)
				{
				fprintf(stderr, "pantelmail: could not allocate space for subject, aborting.\n");	
				}
			break;
		case 'd':
			debug_flag = 1;
			break;
		case 'f':
			free(sender);
			sender = strsave(optarg);	
			if(! sender)
				{
				fprintf(stderr, "pantelmail: could not allocate space for sender, aborting.\n");

				exit(1);
				}
			break;
		case 'h': // help
			print_usage();
			exit(1);
			break;
		case 'p':
			mail_server_port = atoi(optarg);
			break;
		case 'r':
			free(recipient);
			recipient = strsave(optarg);
			if(! recipient)
				{
				fprintf(stderr, "pantelmail: could not allocate space for recipient, aborting.\n");

				exit(1);
				}
			break;
		case 's':
			free(mail_server_name);
			mail_server_name = strsave(optarg);
			if(! mail_server_name)
				{
				fprintf(stderr, "pantelmail: could not allocate space for mail_server_name, aborting.\n");

				exit(1);
				}
			break;
		case 't':
			free(text_filename);
			text_filename = strsave(optarg);
			if(! text_filename)
				{
				fprintf(stderr, "pantelmail: could not allocate space for text_filename, aborting.\n");

				exit(1);
				}			
			break;
		case -1:
        	break;
		case '?':
			if (isprint(optopt) )
 				{
 				fprintf(stderr, "pantelmail: Unknown option `-%c'.\n", optopt);
 				}
			else
				{
				fprintf(stderr, "pantelmail: Unknown option character `\\x%x'.\n", optopt);
				}
			print_usage();

			exit(1);
			break;			
		default:
			print_usage();

			exit(1);
			break;
		}/* end switch a */
	}/* end while getopt() */

/* id */
fprintf(stderr, "Panteltje (c) pantelmail-%s, type pantelmail -h for help.\n", VERSION);

error_flag = 0;
if(strcmp(recipient, "") == 0)
	{
	fprintf(stderr, "pantelmail: no recipient specified, use -r command line flag.\n");

	error_flag = 1;
	}

if(strcmp(subject, "") == 0)
	{
	fprintf(stderr, "pantelmail: no subject specified, use -a command line flag.\n");

	error_flag = 1;
	}

if(strcmp(text_filename, "") == 0)
	{
	fprintf(stderr, "pantelmail: no text file specified, use -t command line flag.\n");

	error_flag = 1;
	}
else
	{
	/* open the text file first */
	text_fileptr = fopen(text_filename, "r");
	if(! text_fileptr)
		{
		fprintf(stderr, "pantelmail: could not open file %s, as specified by the -t command line flag, for read.\n", text_filename);

		error_flag = 1;
		}
	}
	
/* test if no From specified, and no domainname set on system */
sprintf(temp, "%s@", user_name);	
if( (strcmp(domainname, "") == 0) && (strcmp(sender, temp) == 0) )
	{
	fprintf(stderr, "pantelmail: no domainname set, perhaps use 'domainname xxx' to set it, or specify full email address with -f command line flag\n"); 

	error_flag = 1;
	}

if(error_flag)
	{
	fprintf(stderr, "pantelmail: aborting.\n");
	
	exit(1);
	}

if(strcmp(mail_server_name, "") == 0) /* not specified on command line */
	{
	/* extract mail_server_name from recipient */

	if(debug_flag)
		{
		fprintf(stderr, "recipient=%s\n", recipient);
		}

	char *ptr;
	ptr = strstr(recipient, "@");
	if(! ptr)
		{
		fprintf(stderr, "pantelmail: could not extract server from recipient %s, aborting.\n", recipient);

		exit(1);
		}

	free(mail_server_name);
	mail_server_name = strsave(ptr + 1);
	if(!mail_server_name)
		{
		fprintf(stderr, "pantelmail: could not allocate space for mail_server_name, aborting.\n");

		exit(1);
		}

//	if(debug_flag)
		{
		fprintf(stderr, "mail_server_name=%s\n", mail_server_name);
		}


	FILE *pptr;
	char dummy[4096];
	int mx_priority;
	int old_mx_priority;
	char mx_name[4096];
	char mx_final_name[4096];

	/* no host lookup for localhost */
	if(strcmp(mail_server_name, "localhost") != 0)
		{
		sprintf(temp, "host %s", mail_server_name);
		pptr = popen(temp, "r");

		/*
		grml: # host panteltje.com
		panteltje.com has address 81.207.135.196
		panteltje.com mail is handled by 10 ip51cf87c4.direct-adsl.nl.
		panteltje.com mail is handled by 20 mailrelay.direct-adsl.nl.
		*/

		/* get any MX records */
		old_mx_priority = -1;
		while(1)
			{
			a = readline(pptr, temp);	// closes file on EOF flcose? versus pclose
	
			fprintf(stderr, "%s\n", temp);
			if(a == EOF) break;

			/* panteltje.com mail is handled by 20 mailrelay.direct-adsl.nl. */

			a = sscanf(temp, "%s mail is handled by %d %s", dummy, &mx_priority, mx_name);
			if(a == 3)
				{
				fprintf(stderr, "mx_priority=%d mx_name=%s\n", mx_priority, mx_name);

				if(mx_priority > old_mx_priority)
					{
					old_mx_priority = mx_priority;
			
					strcpy(mx_final_name, mx_name);

					/* remove the last dot */
					mx_final_name[strlen(mx_final_name) - 1] = 0;
					}
				}
			}

//		pclose(pptr);

		if(debug_flag)
			{
			fprintf(stderr, "mx_final_name=%s\n", mx_final_name);
			}
		
		free(mail_server_name);
		mail_server_name = strsave(mx_final_name);
		if(! mail_server_name)
			{
			fprintf(stderr, "pantelmail: could not allocate space for mail_server_name (in mx record), abortiong.\n");

			exit(1);
			}

		} /* end if not localhost */
	} /* end if mail_server_name not specified on command line */

/* connect to host or first MX record */

if(! connect_to_mail_server(mail_server_name, mail_server_port) )
	{
	fprintf(stderr, "pantelmail: could not connect to mail server %s port %d, aborting.\n", mail_server_name, mail_server_port);

	exit(1);
	}

fprintf(stderr, "pantelmail: connected to mail server %s port %d\n", mail_server_name, mail_server_port);

/* do the select thing here */
while(1)
	{
	/* this macro initializes the file descriptor readfs to to be the empty set */
	FD_ZERO(&readfs);

	/* this macro adds socketfd to the file descriptor readfs */
	if(mail_socketfd >= 0) FD_SET(mail_socketfd, &readfs);

	timeout.tv_sec = 10;
	timeout.tv_usec = 0;
	/* int FD_SETSIZE  macro is maximum number of filedescriptors that fd_set can hold */
 	/* function select waits for specified filedescr. to have a signal */
	/* last argument struct timeval *timeout */
	b = select(FD_SETSIZE, &readfs, NULL, NULL, &timeout);
	if(b == -1)/* 0 is timeout, -1 error (in errno) */
		{
		fprintf(stderr, "select: failed, error: ");
		if(errno == EBADF) fprintf(stderr, "One of the file descriptor sets specified an invalid file descriptor.\n");
		if(errno == EINTR) fprintf(stderr, "The operation was interrupted by a signal.\n");
		if(errno == EINVAL) fprintf(stderr, "The TIMEOUT argument is invalid; one of the components is negative or too large.\n");
		break;
		}

	/* test if socket has data */
	if( FD_ISSET(mail_socketfd, &readfs) )
		{	
		/* read from mail server */
		mail_io_read_cb(mail_socketfd);
		}

	} /* end while */


exit(1);
} /* end function main */


/*         
211 System status, or system help reply
214 Help message
     [Information on how to use the receiver or the meaning of a
     particular non-standard command; this reply is useful only
     to the human user]
220 <domain> Service ready
221 <domain> Service closing transmission channel
250 Requested mail action okay, completed
251 User not local; will forward to <forward-path>
 
354 Start mail input; end with <CRLF>.<CRLF>
 
421 <domain> Service not available,
     closing transmission channel
     [This may be a reply to any command if the service knows it
     must shut down]
450 Requested mail action not taken: mailbox unavailable
     [E.g., mailbox busy]
451 Requested action aborted: local error in processing
452 Requested action not taken: insufficient system storage
 
500 Syntax error, command unrecognized
     [This may include errors such as command line too long]
501 Syntax error in parameters or arguments
502 Command not implemented
503 Bad sequence of commands
504 Command parameter not implemented
550 Requested action not taken: mailbox unavailable
     [E.g., mailbox not found, no access]
551 User not local; please try <forward-path>
552 Requested mail action aborted: exceeded storage allocation
553 Requested action not taken: mailbox name not allowed
     [E.g., mailbox syntax incorrect]
554 Transaction failed
*/

void mail_io_read_cb(int socket)
{
char rxbuf[READSIZE];
int a, i;
long li;
char c;
char temp[READSIZE];/* beware of overflow rxbuf=512 + additional text */
char temp2[READSIZE];
char arg0[TEMP_SIZE];
static int num_response;/* preserve between calls !!!!! */
int cr_flag;
char *ptr;

if(debug_flag)
	{
	fprintf(stderr, "mail_io_read_cb(): socket=%d\n", socket);
	}

cr_flag = 0;
/* read from socket */
li = 0;
while(1)
	{
	errno = 0;
	a = read(socket, &c, 1);
	if(a < 0)
		{
		if(errno == EAGAIN) continue;/* try again */
		
		if(debug_flag)
			{
			fprintf(stderr, "\npantelmail: mail_io_read_cb(): read failed, connection to server lost errno=%d\n", errno);
			}

		perror("read failed: connection to server lost");

		mail_io_cb_error_return("read failed, connection to mail server lost");
		}
	if(a == 0)
		{
		fprintf(stderr, "pantelmail: mail_io_read_cb(): connection closed by server (EOF)");


		/* close the server socket */
		close(socket);

		mail_server_status = DISCONNECTED;

		command_in_progress = 0;
		return;
		}

	/* valid byte received */

	bytes_received++;
	
	if(c == 10)/* LF */
		{
		if(cr_flag)
			{
			break;
			}
		continue;/* no store of LF */
		}
	
	if(c == 13)/* CR */
		{
		cr_flag = 1;
		continue;/* no store of CR */
		}
	else cr_flag = 0;	

	rxbuf[li] = c;
	li++;
	if(li >= READSIZE) break;
	}/* end while read all characters */
rxbuf[li] = 0;/* string terminator */
/* It is guaranteed that there are no 13 or 10 in the string */

if(debug_flag)
	{
	fprintf(stderr, "pantelmail: mail_io_read_cb(): rxbuf=%s\n", rxbuf);
	}

/* process data */

/* get the numeric response if any */

arg0[0] = 0;
				
sscanf(rxbuf, "%s", arg0);
if( (! isdigit(arg0[0]) ) || (! isdigit(arg0[1]) ) || (! isdigit(arg0[2]) ) )
	{
	if(debug_flag)
		{
		fprintf(stderr, "pantelmail: mail_io_read_cb(): EN: An online error occured: expecting numeric but rxbuf=%s\n", rxbuf);
		}
				
	mail_io_cb_error_return("no numeric responce from mail server");
	return;
	}

if(rxbuf[3] == '-') return; /* 220-something, multi line replies */

num_response = atoi(arg0);

/* process the different commands */
switch(num_response)
	{
	case 211:	/* System status, or system help reply */
		return; /* ignore */
		break;
	case 214:	/*
				Help message
				[Information on how to use the receiver or the meaning of
				a particular non-standard command; this reply is useful
				only to the human user]
				*/
		return; /* ignore */
		break;

	case 220:	/* 220 <domain> Service ready */
		/*
		220-panteltje.pi.net Sendmail 8.6.12/8.6.9 ready at Wed, 4 Feb 1998 16:20:54 +0100
		220 ESMTP spoken here
		*/	
			
		if(mail_mode == 0)
			{
			sprintf(temp, "HELO %s\n", domainname);
			if(! send_to_mail_server(temp) )
				{
				mail_io_cb_error_return("could not send to mail server");
				}
			mail_mode = HELO_SEND;
			return;
			}
		return;
		break;
	case 221:	/* <domain> Service closing transmission channel. Also here as response to QUIT if send was succesful */
		mail_io_cb_error_return("221 <domain> Service closing transmission channel");
		break;
	case 250:	/* repond to helo or mail or message send (\n.\n) */
				/* 250 Requested mail action okay, completed */
		
		if(mail_mode == MESSAGE_SEND)
			{
			if(! send_to_mail_server("QUIT\n") )
				{
				mail_io_cb_error_return("could not send to mail server");
				}


			fprintf(stderr, "pantelmail: email send.\n");

			exit(0);
			}

// NO break!!!
	case 251:	/* User not local; will forward to <forward-path> */
		if(mail_mode == HELO_SEND)
			{
			fprintf(stderr, "pantelmail: mailing %s from %s to %s DO NOT QUIT\n", text_filename, sender, recipient);

			/* here we need to extract the real email address from the sender */
			ptr = strstr(sender, "<");
			if(! ptr)
				{
				fprintf(stderr, "pantelmail: sender not in correct format, needs to be 'name surname <me@mydomain>', aborting.\n");

				exit(1);
				}
			
			sprintf(temp2, "%s", ptr + 1);

			/* get rid of last '>' */
			temp2[strlen(temp2) - 1] = 0;

			sprintf(temp, "MAIL From: %s\n", temp2);
			if(!send_to_mail_server(temp) )
				{
				mail_io_cb_error_return("cannot send to mail server");
				}
			mail_mode = FROM_SEND;
			return;
			}
		if(mail_mode == FROM_SEND)
			{
			if(! recipient)
				{
				mail_io_cb_error_return("cannot read recipient from header file");
				}

			/* extract email field (containing '@') */
			ptr = strstr(recipient, "@");
			if(! ptr) mail_io_cb_error_return("Invalid email address");

			/* move backwards until space or < or start */
			while(1)
				{
				if(ptr == recipient)
					{
					break;/* start found, @ in first field */
					}
				if(*ptr == ' ')
					{
					ptr++;/* point to first valid char in field */
					break;
					}
				/* detect '<' will add later */
				if(*ptr == '<')
					{
					ptr++;/* point to first valid char in field */
					break;
					}
				ptr--;
				}/* end while move backwards */
			/* ptr points to first char in field */

			/* copy field until space or > or end */			
			for(i = 0; i < READSIZE; i++)
				{
				/* detect space */
				if(*ptr == ' ')
					{
					temp[i] = 0;
					break;
					}

				/* detect '>', will add later */
				if(*ptr == '>')
					{
					temp[i] = 0;
					break;
					}
				
				temp[i] = *ptr;

				/* detect end */
				if(*ptr == 0)
					{
					break;
					}
				
				ptr++;
				}/* end while move backwards */

			/*
			now  Jan Panteltje <root@localhost>  will be converted to:
			root@localhost  in string temp.
			*/
			
//			fprintf(stderr, "mailing %ld to %s DO NOT QUIT\n", mail_posting, temp);





			sprintf(temp2, "RCPT To: <%s>\n", temp);/*recipient);*/
			if(!send_to_mail_server(temp2) )
				{
				mail_io_cb_error_return("cannot send to mail server");
				}
			mail_mode = RCPT_SEND;
			return;
			}	
		if(mail_mode == RCPT_SEND)	
			{
			sprintf(temp, "DATA\n");
			if(!send_to_mail_server(temp) )
				{
				mail_io_cb_error_return("cannot send to mail server");
				}
			mail_mode = DATA_SEND;
			return;
			}		
		if(mail_mode == MESSAGE_SEND)
			{
			if(! send_to_mail_server("QUIT\n") ) 
				{
				mail_io_cb_error_return("cannot send to mail server");
				}
			mail_mode = QUIT_SEND;

			close(mail_socketfd);

			/* do not hang around to wait for confirmation of quit */
			mail_server_status = DISCONNECTED;
			
//			set_send_status(mail_posting, SEND_OK);
				
			return;
			}
		break;
	case 354:	/* 354 Start mail input; end with <CRLF>.<CRLF> */
		if(mail_mode != DATA_SEND)
			{
			/* problem */
			mail_io_cb_error_return("num responce 354, but data command was not send");
			return;
			}

		/* write a header */
		/*
		"From: Jan Panteltje <nokia@panteltje.com>"
		"To: +31611495421@gin.nl"
		"Subject: $1"
		"Date: $DATE"
		"Mime-Version: 1.0"
		"Content-Transfer-Encoding: 8bit"
		"Content-Type: text/plain; charset=ISO-8859-15"
		*/		
		
		sprintf(temp2, "From: %s\n", sender);
		if(! send_to_mail_server(temp2) )
			{
			mail_io_cb_error_return("cannot send to mail server");
			return;
			}	

		/* here we need to extract the real email address from the recipient */
		ptr = strstr(recipient, "<");
		if(! ptr)
			{
			fprintf(stderr, "pantelmail: recipient not in correct format, needs to be 'name surname <him@hisdomain>', aborting.\n");

			exit(1);
			}
			
		sprintf(temp2, "%s", ptr + 1);

		/* get rid of last '>' */
		temp2[strlen(temp2) - 1] = 0;

	
		sprintf(temp, "To: %s\n", recipient); 
		if(! send_to_mail_server(temp) )
			{
			mail_io_cb_error_return("cannot send to mail server");
			return;
			}	
	
		sprintf(temp2, "Subject: %s\n", subject); 
		if(! send_to_mail_server(temp2) )
			{
			mail_io_cb_error_return("cannot send to mail server");
			return;
			}	
	
		sprintf(temp2, "Date: %s\n", get_universal_time_as_string() ); 
		if(! send_to_mail_server(temp2) )
			{
			mail_io_cb_error_return("cannot send to mail server");
			return;
			}	
	
		sprintf(temp2, "Mime-Version: 1.0\n");
		if(! send_to_mail_server(temp2) )
			{
			mail_io_cb_error_return("cannot send to mail server");
			return;
			}	
	
		sprintf(temp2, "Content-Transfer-Encoding: 8bit\n");
		if(! send_to_mail_server(temp2) )
			{
			mail_io_cb_error_return("cannot send to mail server");
			return;
			}	
	
		sprintf(temp2, "Content-Type: text/plain; charset=ISO-8859-15"); 
		if(! send_to_mail_server(temp2) )
			{
			mail_io_cb_error_return("cannot send to mail server");
			return;
			}	
	
		/* end header */
			
		/* a blank line separates header and body */
		if(!send_to_mail_server("\n") )
			{
			mail_io_cb_error_return("cannot send to mail server");
			return;
			}

		/* write the message text */
		while(1)
			{
			a = readline(text_fileptr, temp); /* closes file if done */
			strcat(temp, "\n");
			if(! send_to_mail_server(temp) )
				{
				mail_io_cb_error_return("cannot send to mail server");
				return;
				}	
			if(a == EOF) break;
			}

		/* terminate message */	

		sprintf(temp2, ".\n"); 
		if(! send_to_mail_server(temp2) )
			{
			mail_io_cb_error_return("cannot send to mail server");
			return;
			}	

		/* test if there is an attachment */
		/* not implemented, use mpack / munpack in the text file */

		/* 
		Note: last line attachment not guaranteed to have a \n, up to
		mpack
		*/

		/* a single period on a line indicates end of text */
		if(!send_to_mail_server(".\n") )
			{
			mail_io_cb_error_return("cannot send to mail server");
			return;
			}

		mail_mode = MESSAGE_SEND;
			
		return;
		break;

	case 421:	/*
				<domain> Service not available,
			   	closing transmission channel
			   	[This may be a reply to any command if the service knows
			   	it must shut down]
			   	*/
		mail_io_cb_error_return(rxbuf);
		break; 
	case 450:	/*
				Requested mail action not taken: mailbox unavailable
				[E.g., mailbox busy]
				*/
		mail_io_cb_error_return(rxbuf);
		break;		
	case 451:	/* Requested action aborted: local error in processing */
		mail_io_cb_error_return(rxbuf);
		break;
	case 452:	/*
				Requested action not taken: insufficient system storage
 				*/
 		mail_io_cb_error_return(rxbuf);
 		break;
	case 500:	/*
				Syntax error, command unrecognized
				[This may include errors such as command line too long]
				*/
		mail_io_cb_error_return(rxbuf);
		break;		
	case 501:	/* Syntax error in parameters or arguments */
		mail_io_cb_error_return(rxbuf);
		break;
	case 502:	/* Command not implemented */
		mail_io_cb_error_return(rxbuf);
		break;
	case 503:	/* Bad sequence of commands */
		mail_io_cb_error_return(rxbuf);
		break;
	case 504:	/* Command parameter not implemented */
		mail_io_cb_error_return(rxbuf);
		break;
	case 550:	/*
				Requested action not taken: mailbox unavailable
				[E.g., mailbox not found, no access]
				*/
		mail_io_cb_error_return(rxbuf);
		break;
	case 551:	/* User not local; please try <forward-path> */
		mail_io_cb_error_return(rxbuf);
		break;
	case 552:	/*
				Requested mail action aborted: exceeded storage
				allocation
				*/
		mail_io_cb_error_return(rxbuf);
		break;
	case 553:	/*
				Requested action not taken: mailbox name not allowed
			   	[E.g., mailbox syntax incorrect]
				*/
		mail_io_cb_error_return(rxbuf);
		break;
	case 554:	/* Transaction failed */
		mail_io_cb_error_return(rxbuf);
		break;
	default:
		mail_io_cb_error_return(rxbuf);
		break;
	}/* end switch num_response */	

}/* end function mail_io_read_cb */



int mail_io_cb_error_return(char *error_message)
{
if(debug_flag)
	{
	fprintf(stderr, "mail_io_cb_error_return(): arg error_message=%s\n", error_message);
	}

fprintf(stderr, "pantelmail: server says: %s\n", error_message);

/* argument check */
if(! error_message) return 0;

send_to_mail_server("QUIT\n");
mail_mode = QUIT_SEND;

close(mail_socketfd);

/* do not hang around to wait for confirmation of quit */
mail_server_status = DISCONNECTED;


fprintf(stderr, "pantelmail: could not send email.\n");

exit(1);

} /* end function mail_io_cb_error_return */





